<?php
namespace Service;

use Doctrine\ORM\EntityManager;
use Factory\Server as ServerFactory;
use Service\Product as ProductService;
use Zeedhi\DevOps\Model\Configurationmanager\Entities;
use Zeedhi\DevOps\Model\Servermanager\Entities\Server as ServerEntity;
use Zeedhi\DevOps\Model\Servermanager\Entities\ProductServer;
use Zeedhi\DevOps\Util\Environment;

use Util\InstanceProvider;
use Zeedhi\DevOps\Util\Utilities;
use Zeedhi\Framework\Log\Exception;
use Util\Messages;

class Server {
    
    const REMOVE_PRODUCT_SERVICE_ERROR_SUBRELEASE = 'SUBRELEASE_PRODUCT_SERVER_FK';
    const REMOVE_PRODUCT_SERVICE_ERROR_RELEASE = 'RELEASE_PRODUCT_SERVER_FK';
    const REMOVE_PRODUCT_SERVICE_ERROR_DEPLOY = 'DEPLOY_SERVER_PROD_SERVER_FK';

    /** @var EntityManager $entityManager */
    protected $entityManager;
    /** @var ServerFactory $serverFactory */
    protected $serverFactory;
    /** @var \Zeedhi\DevOps\Model\Servermanager\Repositories\Server   */
    protected $serverRepository;
    /** @var \Zeedhi\DevOps\Model\Configurationmanager\Repositories\Subrelease  */
    protected $subReleaseRepository;
    /** @var \Zeedhi\DevOps\Model\Configurationmanager\Repositories\ProductFiles */
    protected $productFilesRepository;
    /** @var \Zeedhi\DevOps\Model\Configurationmanager\Repositories\Release   */
    protected $releaseRepository;
    /** @var \Zeedhi\DevOps\Model\Configurationmanager\Repositories\Deploy */
    protected $deployRepository;
    /** @var \Zeedhi\DevOps\Model\Servermanager\Repositories\DeployServer */
    protected $deployServerRepository;
    /** @var \Zeedhi\DevOps\Model\Devops\Repositories\ProductFilesMnemonic */
    protected $serverMnemonicRepository;
    /** @var \Zeedhi\DevOps\Model\Projectmanagement\Repositories\Resource */
    protected $resourceRepository;
    /** @var Environment */
    private $environment;
    /**
     * @var Mnemonic
     */
    private $mnemonicService;
    /**
     * @var ProductService
     */
    protected $productService;
    /**
     * @param EntityManager $entityManager
     * @param Environment $environment
     * @param ServerFactory $serverFactory
     * @param Mnemonic $mnemonicService
     * @param ProductService $productService
     */
    public function __construct(EntityManager $entityManager, Environment $environment, ServerFactory $serverFactory,
                                Mnemonic $mnemonicService, ProductService $productService) {
        $this->environment = $environment;
        $this->entityManager = $entityManager;
        $this->serverFactory = $serverFactory;
        $this->productService = $productService;
        $this->mnemonicService = $mnemonicService;
        $this->mnemonicRepository = $this->entityManager->getRepository(Utilities::DEVOPS_ENTITY_DIR . 'Mnemonic');
        $this->serverRepository = $this->entityManager->getRepository(Utilities::SERVER_MANAGER_ENTITY_DIR . 'Server');
        $this->deployRepository = $this->entityManager->getRepository(Utilities::CONFIGURATION_MANAGER_ENTITY_DIR . 'Deploy');
        $this->serverMnemonicRepository = $this->entityManager->getRepository(Utilities::DEVOPS_ENTITY_DIR . 'ServerMnemonic');
        $this->releaseRepository = $this->entityManager->getRepository(Utilities::CONFIGURATION_MANAGER_ENTITY_DIR . 'Release');
        $this->deployServerRepository = $this->entityManager->getRepository(Utilities::SERVER_MANAGER_ENTITY_DIR . 'DeployServer');
        $this->subReleaseRepository = $this->entityManager->getRepository(Utilities::CONFIGURATION_MANAGER_ENTITY_DIR . 'Subrelease');
        $this->productFilesRepository = $this->entityManager->getRepository(Utilities::CONFIGURATION_MANAGER_ENTITY_DIR . 'ProductFiles');
        $this->resourceRepository = $this->entityManager->getRepository(Utilities::PROJECT_MANAGEMENT_ENTITY_DIR . 'Resource');
    }

    /**
     * @param $productServerId - Id da associação entre Servidor/Produto
     * @param $serverId - Id do Servidor
     * @param $productId - Id do Produto
     * @param $name - Nome do Servidor
     * @param $wsEmail - Email do WebSocket
     * @param $useType - Tipo de uso do Servidor (Teste, Produção...)
     * @param $description - Descrição do Servidor
     * @param $baseUrl - Url base da associação entre Servidor/Produto
     * @param $basePath - Diretório base da associação entre Servidor/Produto
     * @param $id
     * @param null $backupLimit
     * @param string $onlyApproved
     * @param string $isActive
     * @param null $resourceId
     * @throws \Exception
     * @return int
     */
    public function insertOrUpdateServer($productServerId, $serverId, $productId, $name, $wsEmail, $useType, $description,
                                         $baseUrl, $basePath, $id, $backupLimit = null, $onlyApproved = 'Y', $isActive = 'Y',
                                         $resourceId = null, $isProductServer = null, $isWindows = null, $isProduction = null) {
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        try {
            if($serverId) {
                $server = $this->entityManager->find(Utilities::SERVER_MANAGER_ENTITY_DIR . 'Server', $serverId);
            } else if($id) {
                $server = $this->entityManager->find(Utilities::SERVER_MANAGER_ENTITY_DIR . 'Server', $id);
            } else {
                $server = new ServerEntity();
            }

            /** @var \Zeedhi\DevOps\Model\Servermanager\Entities\Server $server */
            if(!$productId) {
                $resource = null;
                if($resourceId) {
                    /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Resource $resource */
                    $resource = $this->resourceRepository->find($resourceId);
                }
                $server = $this->serverFactory->setServerRecords($server, $name, $wsEmail, $description, $backupLimit,
                    $onlyApproved, $resource, $isWindows, $isActive, $isProduction);
            }
            $this->entityManager->persist($server);
            $this->entityManager->flush();

            if($productId) {
                /** @var \Zeedhi\DevOps\Model\Configurationmanager\Entities\Product $server */
                $product = $this->entityManager->find(Utilities::CONFIGURATION_MANAGER_ENTITY_DIR . 'Product', $productId);

                if($productServerId) {
                    /** @var \Zeedhi\DevOps\Model\Servermanager\Entities\ProductServer $productServer */
                    $productServer = $this->entityManager->find(Utilities::SERVER_MANAGER_ENTITY_DIR . 'ProductServer', $productServerId);

                    if (!is_object($productServer)) {
                        $productServer = new ProductServer();
                    }
                } else {
                    $productServer = new ProductServer();
                }

                $productServer = $this->serverFactory->setProductServerRecords($productServer, $product, $server, $useType, $baseUrl, $basePath, $backupLimit, $isActive);
                $this->entityManager->persist($productServer);
                $this->entityManager->flush();
            }

            $connection->commit();
            return ($productId) ? $productServer->getId() : $server->getId();
            
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

    public function insertOrUpdateServerForEach($serverData){
            $products = $serverData['PRODUCTS'];
            $productServer = array();
            foreach($products as $productId){
                $baseUrl         = isset($serverData['BASE_URL_' . $productId]) ? $serverData['BASE_URL_' . $productId] : null;
                $basePath        = isset($serverData['BASE_PATH_' . $productId]) ? $serverData['BASE_PATH_' . $productId] : null;
                $backupLimit     = $serverData['BACKUP_LIMIT_' . $productId];
                $name            = $serverData['PRODUCT_NAME_' . $productId];
                $serverId        = $serverData['SERVER_ID'];
                $useType         = isset($serverData['USETYPE']) ? $serverData['USETYPE'] : null;
                $productServer[] = $this->insertOrUpdateServer(null, $serverId, $productId, $name, null, $useType, null, $baseUrl, $basePath, null, $backupLimit);
            }
            return $productServer;
    }

    /**
     * @param $serverId - Id do Servidor
     * @return array
     * @throws \Exception
     *
     * Apaga a associação entre Servidor/Produto e o próprio Servidor, se possível.
     */
    public function removeServer($serverId){
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        try {
            $productNames = array();

            /** @var \Zeedhi\DevOps\Model\Servermanager\Entities\Server $server */
            $server = $this->entityManager->find(Utilities::SERVER_MANAGER_ENTITY_DIR . 'Server', $serverId);

            $allProductServers = $this->entityManager->getRepository(Utilities::SERVER_MANAGER_ENTITY_DIR . 'ProductServer')
                ->findBy(array('server' => $server));

            if(empty($allProductServers)) {
                $this->entityManager->remove($server);
            } else {
                foreach($allProductServers as $prodServer) {
                    /** @var \Zeedhi\DevOps\Model\Configurationmanager\Entities\Product $server */
                    $product = $this->entityManager->find(Utilities::CONFIGURATION_MANAGER_ENTITY_DIR. 'Product', $prodServer->getProduct()->getId());

                    $productNames[] = '"' . $product->getName() . '"';
                }
            }

            $this->entityManager->flush();
            $connection->commit();
            return $productNames;
        }
        catch(\Exception $e){
            $connection->rollBack();
            throw $e;
        }
    }
    
    public function findProductsByServer($serverId){
        return $this->productService->findProductsByServer($serverId);
    }

    /**
     * @param $id
     * @throws \Doctrine\DBAL\ConnectionException
     * @throws \Exception
     *
     * Apaga a associação entre Servidor/Produto.
     */
    public function removeProductServer($id){
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        try {
            /** @var \Zeedhi\DevOps\Model\Servermanager\Entities\ProductServer $productServer */
            $productServer = $this->entityManager->find(Utilities::SERVER_MANAGER_ENTITY_DIR . 'ProductServer', $id);

            if(is_object($productServer)) {
                $this->entityManager->remove($productServer);
            }

            $this->entityManager->flush();
            $connection->commit();
        }
        catch(\Exception $e){
            $connection->rollBack();
            throw new Exception($this->getErrorHandledWhenDeletingProductServer($e->getMessage()));
        }
    }
    
    public function getErrorHandledWhenDeletingProductServer($errorMessage) {
        if(strpos($errorMessage, 'FK') && strpos($errorMessage,'ORA-02292')) {
            $initPointString    = strpos($errorMessage, '.') + 1;
            $errorStringPartial = substr($errorMessage, $initPointString);
            $endPointString     = strpos($errorStringPartial, ')');
            $stringError        = substr($errorStringPartial, 0, $endPointString);
            if($stringError == self::REMOVE_PRODUCT_SERVICE_ERROR_SUBRELEASE){
                return Messages::ERR_PRODUCT_SERVER_CANNOT_BE_DELETED . 'Sub-Release';
            } else if($stringError == self::REMOVE_PRODUCT_SERVICE_ERROR_RELEASE){
                return Messages::ERR_PRODUCT_SERVER_CANNOT_BE_DELETED . 'Release';
            } else if($stringError == self::REMOVE_PRODUCT_SERVICE_ERROR_DEPLOY) {
                return Messages::ERR_PRODUCT_SERVER_CANNOT_BE_DELETED . 'Deploy';
            } else {
                return $errorMessage;
            }
        } else {
            return $errorMessage;
        }
        
    }
    

    /**
     * @param $productId - Id do Produto
     * @return array
     *
     * Retorna os servidores de um produto.
     */
    public function findServersWithFilter($filters, $page, $pageSize){
        $organizationId = $this->environment->getOrganizationId();
        $limit = $page * $pageSize;
        $offset = ($page - 1) * $pageSize;
        return $this->serverRepository->findServersWithFilter($filters, $organizationId, $limit, $offset);
    }

    public function findMnemonicByServer($serverId, $filters, $filterWhereInByProductsId, $page = null, $pageSize = null){
        $limit = $page * $pageSize;
        $offset = ($page - 1) * $pageSize;
        return $this->mnemonicService->findMnemonicByServer($serverId, $filters, $filterWhereInByProductsId, $limit, $offset);

    }
    public function findMnemonicToExport($serverId, $exportedServerRows){
        $exportedMnemonics = $this->mnemonicRepository->searchMnemonicByServerId($serverId);
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
         $serverMnemonics = [];
         
        try{
                foreach($exportedServerRows as $exportedServerRow){
                    
                    $exportedServerId = $exportedServerRow['ID'];
                    if($exportedServerId !== $serverId){
                        $mnemonics = $this->mnemonicRepository->searchMnemonicIdByserver($exportedServerId);
                        foreach($mnemonics as $mnemonic){
                            array_push($serverMnemonics, $mnemonic['MNEMONIC_ID']);
                        } 
                        foreach($exportedMnemonics as $exportedMnemonic){
                            if(!in_array($exportedMnemonic['MNEMONIC_ID'],  $serverMnemonics)){
                                $mnemonicEntity = $this->mnemonicRepository->find($exportedMnemonic['MNEMONIC_ID']);
                                $server = $this->serverRepository->find($exportedServerId);
                                $serverMnemonic = $this->mnemonicService->createServerMnemonic($mnemonicEntity, $server, $exportedMnemonic['VALUE']);
                                $this->entityManager->persist($serverMnemonic);
                                $this->entityManager->flush();
                            }else{
                                $mnemonicEntity = $this->serverMnemonicRepository->findOneBy(array('server' => $exportedServerId, 'mnemonic' =>$exportedMnemonic['MNEMONIC_ID']));
                                if($mnemonicEntity){
                                    $serverMnemonic = $this->mnemonicService->updateServerMnemonic($mnemonicEntity, $exportedMnemonic['VALUE']);
                                    $this->entityManager->persist($serverMnemonic);
                                    $this->entityManager->flush();
                                }
                                
                            }
                    }
                }
            } 
            $connection->commit();
            return $serverId;
        } catch (\Exception $ex) {
            $connection->rollBack();
            throw $ex;
        }
    }

    public function saveMnemonicByServer($serverId, $mnemonicId, $mnemonicString, $value, $label, $isEncrypted = 'N', $mnemonicGrouping = null)
    {
        
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        try {
            $server = $this->serverRepository->find($serverId);
            $mnemonic = $this->mnemonicRepository->findOneBy(array('id' => $mnemonicId));
            $serverMnemonic = $this->serverMnemonicRepository->findOneBy(array('mnemonic' => $mnemonic, 'server' => $server));
            $oldValue = $serverMnemonic ? $serverMnemonic->getValue() : null;
            if (!empty($value)){
                $idMnemonic = $this->mnemonicService->saveMnemonicByServer($server, $mnemonic, $mnemonicString, $value, $label, $isEncrypted, $mnemonicGrouping);
            } else if (empty($value) && empty($mnemonic->getDefaultValue())) {                
                throw new \Exception(Messages::ERR_MNEMONIC_VALUE_NOT_INFORMED);
            } else if (empty($value) && $serverMnemonic) {                 
                $this->entityManager->remove($serverMnemonic);
                $this->entityManager->flush();
            }
            $serverMnemonic = $this->serverMnemonicRepository->findOneBy(array('mnemonic' => $mnemonic, 'server' => $server));
            $filter = "AND M.MNEMONIC_ID = ". $mnemonic->getId();
            $mnemonicRow = $this->mnemonicService->findMnemonicByServer($server->getId(), $filter)[0];
            
            if($serverMnemonic) {
                $date = new \DateTime();
                $mongoRow = array(
                    'MNEMONIC_SERVER_ID'  => intval($serverMnemonic->getId()),
                    'OLD_VALUE'     => $oldValue,
                    'NEW_VALUE' => $value,
                    'NMOPERADOR'  => $this->environment->getCurrentUserName(),
                    "DATE" => intval($date->format('YmdHis'))
                );
                InstanceProvider::getMongoDB()->insert("MNEMONIC_SERVER_LOG", $mongoRow);
            }
            $connection->commit();
            return $mnemonicRow;
        } catch (\Exception $ex) {
            $connection->rollBack();
            throw $ex;
        }
    }

    public function removeMnemonicByServer($mnemonicId, $serverId)
    {
        $serverMnemonic = $this->serverMnemonicRepository->findOneBy(array('mnemonic' => $mnemonicId,'server' => $serverId));
        if($serverMnemonic){
            $this->entityManager->remove($serverMnemonic);
            $this->entityManager->flush();
        }
    }

    /**
     * @param null $filters
     * @return array
     *
     * Retorna todos os servidores.
     */
    public function findServers($filters, $page = null, $pageSize = null) {
        $organizationId = $this->environment->getOrganizationId();
        $limit = $page * $pageSize;
        $offset = ($page - 1) * $pageSize;
        return $this->serverRepository->findServers($organizationId, $filters, $limit, $offset);
    }

    public function setStateApproved() {
        $organizationId = $this->environment->getOrganizationId();
        return $this->serverRepository->findServers($organizationId);
    }

    /**
     * @param $productId - Id do Produto
     * @param $type - Tipo da relação entre produto e servidor
     * @return array
     *
     * Retorna os servidores de um produto filtrados pelo tipo de associação.
     */
    public function serverByProductAndType($productId, $type, $filterProductServerStatus = null) {
        return $this->serverRepository->findServerByProductAndType($productId, $type, $filterProductServerStatus);
    }

    public static function scripTypes(){
        return array(
            ServerEntity::APPROVED_TEST => array(
                Entities\ProductFiles::TEST_SCRIPT_AFTER,
                Entities\ProductFiles::TEST_SCRIPT_BEFORE,
            ),
            ServerEntity::APPROVED_PACKAGE => array(
                Entities\ProductFiles::PACKAGE_SCRIPT_AFTER,
                Entities\ProductFiles::PACKAGE_SCRIPT_BEFORE,
                Entities\ProductFiles::HOMOLOGATION_SCRIPT_AFTER,
                Entities\ProductFiles::HOMOLOGATION_SCRIPT_BEFORE,
            ),
            ServerEntity::APPROVED_DEPLOY => array(
                Entities\ProductFiles::DEPLOY_SCRIPT_AFTER,
                Entities\ProductFiles::DEPLOY_SCRIPT_BEFORE,
                Entities\ProductFiles::BACKUP_SCRIPT_AFTER,
                Entities\ProductFiles::BACKUP_SCRIPT_BEFORE,
            ),
            ServerEntity::APPROVED_BACKUP => array(
                Entities\ProductFiles::BACKUP_SCRIPT_AFTER,
                Entities\ProductFiles::BACKUP_SCRIPT_BEFORE,
            ),
            ServerEntity::APPROVED_FINISH => array(
                Entities\ProductFiles::PACKAGE_SCRIPT_AFTER,
                Entities\ProductFiles::PACKAGE_SCRIPT_BEFORE,
            )
        );
    }

    public static function getScripTypesByOperation($index){
        return self::scripTypes()[$index];
    }

    public function validateSubReleaseServerPermissions($subReleaseId, $operation)
    {
        /** @var Entities\Subrelease $subRelease */
        $subRelease = $this->subReleaseRepository->find($subReleaseId);
        $product = $subRelease->getRelease()->getProduct();
        $server = $subRelease->getServerByType($operation);
        $types = self::getScripTypesByOperation($operation);
        $needsApproval = $this->getNeedApproval($product, $server, $types);
        return $needsApproval;
    }

    public function validateReleaseServerPermissions($releaseId, $operation)
    {
        /** @var Entities\Release $release */
        $release = $this->releaseRepository->find($releaseId);
        $product = $release->getProduct();
        $server = $release->getProductServer()->getServer();
        $types = self::getScripTypesByOperation($operation);
        $needsApproval = $this->getNeedApproval($product, $server, $types);
        return $needsApproval;
    }


    public function validateDeployServerPermissions($deployId, $operation, $server = null)
    {
        /** @var Entities\Deploy $deploy */
        $deploy = $this->deployRepository->find($deployId);
        $product = $deploy->getProduct();
        /** @var \Zeedhi\DevOps\Model\Servermanager\Entities\DeployServer[]  $deployServes */
        $deployServes = $this->deployServerRepository->findBy(array('deploy' => $deploy ));
        $needsApproval = array();
        $types = self::getScripTypesByOperation($operation);
        if($server){
            $server  = $this->deployServerRepository->find($server);
            $needsApproval = $this->getNeedApproval($product, $server->getProductServer()->getServer(), $types, $needsApproval);
        }else {
            foreach ($deployServes as $deployServe) {
                $server = $deployServe->getProductServer()->getServer();
                $needsApproval = $this->getNeedApproval($product, $server, $types, $needsApproval);
            }
        }
        return $needsApproval;
    }

    /**
     * @param Entities\Product $product
     * @param ServerEntity $server
     * @param $scripTypes
     * @param array $needsApproval
     * @return array
     */
    protected function getNeedApproval(Entities\Product $product,
                                       ServerEntity $server = null,$scripTypes,$needsApproval = array())
    {

        if ($server && $server->getOnlyApprovedScripts() == 'Y') {
            /** @var Entities\ProductFiles[] $productFiles */
            $productFiles = $this->productFilesRepository
                ->findBy(array(
                        'product' => $product,
                        'type' => $scripTypes
                    )
                );
            foreach ($productFiles as $productFile) {
                if ($productFile->getState() == Entities\ProductFiles::SCRIPT_PENDING) {
                    $script = array('name' => $productFile->getFileName(), 'type' => $productFile->getType());
                    if (!in_array($script, $needsApproval)) {
                        $needsApproval[] = $script;
                    }
                }
            }
            return $needsApproval;

        }
        return $needsApproval;
    }
}
